/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __LIST_SKELETONS_WITH_FUSION_TPP__
#define __LIST_SKELETONS_WITH_FUSION_TPP__

#include "list_skeletons_common.tpp"

#endif // __LIST_SKELETONS_WITH_FUSION_TPP__

namespace sketo {
  namespace _impl {

    template <typename OPLUS, typename AS, typename A>
    A
    ls_adpt::reduce(const OPLUS &oplus, const A &identity, const AS &as)
    {
      as.make_local_get_ready();
      // local reduce
      reduce_comp<OPLUS, A> comp(oplus, identity);
      _impl::ls_adpt::dist_list_driver(comp, as);
      //std::vector<A> &vals = comp.vals;  // this make the above main loop **SLOW** 
      std::vector<A> vals = comp.vals;
      
      // global reduce
      _impl::ls_adpt::global_reduce(oplus, identity, vals.size(), &(*vals.begin()));
      return vals[0];
    }

    template <typename OPLUS, typename A, typename G, bool overwriting = false, bool special_impl = false>
    struct scan_comp {
      _impl::dist_list_buffer<typename G::result_type> *data;
      std::vector<A> &total_sums;
      A total_sum;
      const OPLUS &oplus;
      const A &identity;
      const G &g;
      scan_comp(_impl::dist_list_buffer<typename G::result_type>*data, const OPLUS &oplus, const A &identity, std::vector<A> &total_sums, const G&g) : data(data), oplus(oplus), identity(identity), total_sums(total_sums), g(g) { }
    typename G::result_type*buf;
    inline void init(int l, int ls) {
      if(!overwriting) {
        data->bufs.push_back(BUFFER<typename G::result_type>());
        data->bufs.back().resize(ls);
      }
      buf = &(data->bufs[l][0]);
      total_sum = special_impl ? total_sums[l] : identity;
    }
    inline void receive(int i, const A& a) {
     
      buf[i] = g(total_sum);
      total_sum = oplus(total_sum, a);
    }
    inline void fin(int) {
      if(!special_impl) total_sums.push_back(total_sum);
    }
  };

    template <typename OP, typename AS, typename A, typename G, bool overwriting>
    void
    ls_adpt::scanobj_ow_impl(_impl::dist_list_buffer<typename G::result_type>* data, const OP &oplus, const A &identity, const AS &as, const A &e, A* last_val, const G&g)
    {
      as.make_local_get_ready();
      // local reduce
      reduce_comp<OP, A> comp(oplus, identity);
      _impl::ls_adpt::dist_list_driver(comp, as);
      
      std::vector<A> total_sums = comp.vals;
        
      // global scan
      std::vector<A> first_vals(total_sums.size(), identity);
      global_scan(oplus, identity, total_sums.size(), &(*first_vals.begin()), &(*total_sums.begin()));
      
      for(int l = 0; l < total_sums.size(); l++) {
        if (rank == 0 && l == 0) {
          first_vals[l] = e;
        } else {
          first_vals[l] = oplus(e, first_vals[l]);
        }
      }
    
      // share the last value (= total_sum)
      if (last_val) {
        *last_val = oplus(e, total_sums.back());
      }

      // assumption: bs has the same shapse as as.
      // local scan
      {
        scan_comp<OP, A, G, overwriting, true> comp(data, oplus, identity, first_vals, g);
        if(overwriting) _impl::ls_adpt::dist_list_driver_ow(comp, as);
        else            {
          data->bufs.clear();
          _impl::ls_adpt::dist_list_driver(comp, as);
          data->dirty = as.is_dirty();
          if(!data->dirty) {
            data->sizes = as.get_sizes();
            data->global_size = as.get_global_size();
          }
        }
      }
    }



    template <typename OPLUS, typename A, typename G, bool overwriting = false, bool special_impl = false>
    struct scanr_comp {
      _impl::dist_list_buffer<typename G::result_type> *data;
      std::vector<A> &total_sums;
      A total_sum;
      const OPLUS &oplus;
      const A &identity;
      const G &g;
      scanr_comp(_impl::dist_list_buffer<typename G::result_type>*data, const OPLUS &oplus, const A &identity, std::vector<A> &total_sums, const G&g) : data(data), oplus(oplus), identity(identity), total_sums(total_sums), g(g) { }
      typename G::result_type*buf;
      inline void init(int l, int ls) {
	if(!overwriting) {
	  data->bufs.push_back(BUFFER<typename G::result_type>());
	  data->bufs.back().resize(ls);
	}
        buf = &(data->bufs[l][0]);
        total_sum = special_impl ? total_sums[l] : identity;
      }
      inline void receive(int i, const A& a) {
        buf[i] = g(total_sum);
        total_sum = oplus(a, total_sum);
      }
      inline void fin(int) {
        if(!special_impl) total_sums.push_back(total_sum);
      }
    };
    
    template <typename OP, typename AS, typename A, typename G, bool overwriting>
    void
    ls_adpt::scanrobj_ow_impl(_impl::dist_list_buffer<typename G::result_type>* data, const OP &oplus, const A &identity, const AS &as, const A &e, A* first_val, const G&g)
    {
      // local reduce
      as.make_local_get_ready();
      reduce_comp<OP, A> comp(oplus, identity);
      _impl::ls_adpt::dist_list_driver(comp, as);
      std::vector<A> total_sums = comp.vals;

      // global scan
      std::vector<A> last_vals(total_sums.size(), identity);
      global_scanr(oplus, identity, total_sums.size(), &(*last_vals.begin()), &(*total_sums.begin()));

      for(int l = 0; l < total_sums.size(); l++) {
        if (rank == sketo::procs - 1 && l == total_sums.size() - 1) {
          last_vals[l] = e;
        } else {
          last_vals[l] = oplus(last_vals[l], e);
        }
      }
    
      // share the first value (= total_sum)
      if (first_val) {
        *first_val = oplus(total_sums[0], e);
      }

      // local scan
      {
        scanr_comp<OP, A, G, overwriting, true> comp(data, oplus, identity, last_vals, g);
	if(overwriting) _impl::ls_adpt::dist_list_driver_rev_ow(comp, as);
        else            {
          data->bufs.clear();
          _impl::ls_adpt::dist_list_driver_rev(comp, as);
          data->dirty = as.is_dirty();
          if(!data->dirty) {
            data->sizes = as.get_sizes();
            data->global_size = as.get_global_size();
          }
        }

      }
    
    }

    template <typename A>
    void
    _impl::ls_adpt::shiftl_edge(const std::vector<A>& old_leftmost,
                                std::vector<A>& new_rightmost,
                                const std::vector< std::vector<int> > &sizes)
    {
      std::vector< simple_send_request<A> > reqs;
      std::vector< simple_recv_request<A> > reqr;
      int l, r;
      int pl = -1, pr = -1;
      bool first = true;
      
      for(l = sizes.size() - 1; l >= 0; l--) {
        for(r = sketo::procs - 1; r >= 0; r--) {
          if(sizes[l][r] > 0) {
            if(first) {
              first = false;
              // new_rightmost[l] = fromR;  // done by the caller
            } else {
              if(sketo::rank == r) {  // receiver
                if(sketo::rank == pr) { // self-copy
                  new_rightmost[l] = old_leftmost[pl];
                } else {
                  reqr.push_back(simple_recv_request<A>());
                  _impl::transporter::irecv(&new_rightmost[l], 1, pr,
                                            TAG_SHIFTL, &reqr.back());
                }
              } else if (sketo::rank == pr) { // sender
                reqs.push_back(simple_send_request<A>());
                _impl::transporter::isend(&old_leftmost[pl], 1, r,
                                          TAG_SHIFTL, &reqs.back());
              }
            }
            pl = l;
            pr = r;
          }
        }
      }
      // waiting for the completion of the communication
      for(int i = 0; i < reqs.size(); i++) reqs[i].wait();
      for(int i = 0; i < reqr.size(); i++) reqr[i].wait();

    }

    template <typename A>
    void
    _impl::ls_adpt::shiftr_edge(const std::vector<A>& old_rightmost,
                                std::vector<A>& new_leftmost,
                                const std::vector< std::vector<int> > &sizes)
    {
      std::vector< simple_send_request<A> > reqs;
      std::vector< simple_recv_request<A> > reqr;
      int l, r;
      int pl = -1, pr = -1;
      bool first = true;
      
      for(l = 0; l < sizes.size(); l++) {
        for(r = 0; r < sketo::procs; r++) {
          if(sizes[l][r] > 0) {
            if(first) {
              first = false;
              // new_leftmost[l] = fromL;  // done by the caller
            } else {
              if(sketo::rank == r) {  // receiver
                if(sketo::rank == pr) { // self-copy
                  new_leftmost[l] = old_rightmost[pl];
                } else {
                  reqr.push_back(simple_recv_request<A>());
                  _impl::transporter::irecv(&new_leftmost[l], 1, pr,
                                            TAG_SHIFTR, &reqr.back());
                }
              } else if (sketo::rank == pr) { // sender
                reqs.push_back(simple_send_request<A>());
                _impl::transporter::isend(&old_rightmost[pl], 1, r,
                                          TAG_SHIFTR, &reqs.back());
              }
            }
            pl = l;
            pr = r;
          }
        }
      }
      // waiting for the completion of the communication
      for(int i = 0; i < reqs.size(); i++) reqs[i].wait();
      for(int i = 0; i < reqr.size(); i++) reqr[i].wait();

    }

    template <typename A>
    dist_list_buffer<A> *
    ls_adpt::get_dist_list_buffer(const dist_list<A> &as)
    {
      return as.data;
    }

  template <typename OPLUS, typename A>
  struct reduce_comp : public map_reduce_comp<sketo::functions::identity<A>, OPLUS, A> {
    reduce_comp(const OPLUS &oplus, const A&identity) : map_reduce_comp<sketo::functions::identity<A>, OPLUS, A>(sketo::functions::identity<A>(), oplus, identity) { }
  };

    template <typename F, typename AS, typename B, typename CMOBJ, typename G>
    dist_list<B> 
    ls_adpt::concatmapobj_materialize(const CMOBJ &cmobj)
    {
      // if(sketo::rank==0) std::cout << "materialize!" << std::endl;
      typedef typename AS::InitialType IL;
      typedef typename G::result_type A;
      
      const IL &dl = cmobj.as.getInitialList();
      dl.make_local_get_ready();

      dist_list<B> ret(0);
      
      ret.data->bufs.clear();
      for(int l = 0; l < dl.get_num_bufs(); l++) {
        ret.data->bufs.push_back(BUFFER<B>());
        VectorCollector<B, G> ic(ret.data->bufs.back(), cmobj.g);
        typename CMOBJ::template CollectorType< VectorCollector<B, G> >::type c = cmobj.getCollector(ic);
        for(int i = 0; i < dl.get_local_size(l); i++) {
          c.push_back(dl.local_get(i, l));
        }        
      }
      ret.data->dirty = true;
      return ret;
    }
    
    template <typename OPLUS, typename FF, typename AS, typename G, typename A, typename IL>
    struct reduce_concatmap_comp {
      A val;
      std::vector<A> vals;
      typedef _impl::ReduceCollector<A, OPLUS, G> InitialCollectorType;
      typedef typename ls_concatmapobj<FF, AS, G>::template CollectorType< InitialCollectorType >::type CombinedCollectorType;
      InitialCollectorType ic;
      CombinedCollectorType c;
      const IL& il;
      const OPLUS &oplus;
      const A &identity;
      typedef _impl::ls_concatmapobj<FF, AS, G> CMOBJ;
      const CMOBJ &cmobj;
      reduce_concatmap_comp(const OPLUS &oplus, const A&identity, const CMOBJ &cmobj, const IL &il) : oplus(oplus), identity(identity), cmobj(cmobj), il(il), ic(val, oplus, cmobj.g), c(cmobj.getCollector(ic)) { }
      inline void init(int , int ) {
        val = identity;
      }
      inline void receive(int , const typename IL::element_type& a) {
        c.push_back(a);
      }
      inline void fin(int ) {
        vals.push_back(val);
      }
    };
    /** specialization of reduce for reduce-concatmap fusion */
    template <typename OPLUS, typename FF, typename AS, typename G>
    typename OPLUS::result_type
    ls_adpt::reduce_concatmap(const OPLUS &oplus,
                              const typename OPLUS::result_type &identity,
                              const _impl::ls_concatmapobj<FF, AS, G> &cmobj) {
      //if(sketo::rank==0) std::cout << "reduce!" << std::endl;
      typedef typename AS::InitialType IL;
      typedef typename OPLUS::result_type A;
      
      const IL &dl = cmobj.as.getInitialList();
      dl.make_local_get_ready();

      reduce_concatmap_comp<OPLUS, FF, AS, G, A, IL> comp(oplus, identity, cmobj, dl);
      _impl::ls_adpt::dist_list_driver(comp, dl);
      std::vector<A> vals = comp.vals;

      // std::vector<typename OPLUS::result_type> vals;
      // for(int l = 0; l < dl.get_num_bufs(); l++) {
      //   typename OPLUS::result_type res = identity;
      //   _impl::ReduceCollector<typename G::result_type, OPLUS, G> ic(res, oplus, cmobj.g);
      //   typename ls_concatmapobj<FF, AS, G>::template CollectorType< _impl::ReduceCollector<typename G::result_type, OPLUS, G> >::type c = cmobj.getCollector(ic);
      //   for(int i = 0; i < dl.get_local_size(l); i++) {
      //     c.push_back(dl.local_get(i, l));
      //   }
      //   vals.push_back(res);
      // }
      
      // global reduce
      _impl::ls_adpt::global_reduce(oplus, identity, vals.size(), &(*vals.begin()));
      return vals[0];
      
    }
    
  }

  /************** TENTATIVE. TO BE REMOVED    **************/


  //------------------------------------------------------------------------
  // append
  template <typename A>
  dist_list<A>
  list_skeletons::append(const dist_list<A> &as,
                         const dist_list<A> &bs)
  {
    return _impl::ls_adpt::append(as, bs);
  }

  template <typename A>
  dist_list<A>
  _impl::ls_adpt::append(const dist_list<A> &as,
                         const dist_list<A> &bs)
  {
    dist_list<A> ret(0);
    ret.data->bufs.clear();
    ret.data->bufs.reserve(as.data->bufs.size() + bs.data->bufs.size());
    for(int l = 0; l < as.data->bufs.size(); l++) ret.data->bufs.push_back(as.data->bufs[l]);
    for(int l = 0; l < bs.data->bufs.size(); l++) ret.data->bufs.push_back(bs.data->bufs[l]);
    if(!as.is_dirty() && !bs.is_dirty()) {
      ret.data->dirty = false;
      ret.data->sizes.clear();
      std::vector< std::vector<int> > asizes = as.get_sizes();
      std::vector< std::vector<int> > bsizes = bs.get_sizes();
      for(int i = 0; i < asizes.size(); i++) ret.data->sizes.push_back(asizes[i]);
      for(int i = 0; i < bsizes.size(); i++) ret.data->sizes.push_back(bsizes[i]);
      ret.data->global_size = as.get_global_size() + bs.get_global_size();
    } else {
      ret.data->dirty = true;
    }
    return ret;
  }

}
