#include <iostream>
#include <cmath>
#include <utility>

#include <sketo/list_skeletons.h>
#ifndef GSHIFT_DUMMY
#include "gshift.h"
#else
#include "gshift_dummy.h"
#endif

namespace sl = sketo::list_skeletons;

#include "testcase.h"

struct gen_b_t : public sketo::functions::base<double (int)> {
  double operator()(int m) const {
    int i = m % (imax + 2); m /= (imax + 2);
    int j = m % (jmax + 2); m /= (jmax + 2);
    int k = m;
    if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1) ||
	(k == 0) || (k == kmax + 1)) {
      return 0.0;
    }
    double ret = 2.0;
    if ((i > 1) && (i < imax)) ret += 2.0; else ret += 1.0;
    if ((j > 1) && (j < jmax)) ret += 2.0; else ret += 1.0;
    if ((k > 1) && (k < kmax)) ret += 2.0; else ret += 1.0;
    return ret;
  }
} gen_b;

struct gen_f_t : public sketo::functions::base<double (int)> {
  double operator()(int m) const {
    int i = m % (imax + 2); m /= (imax + 2);
    int j = m % (jmax + 2); m /= (jmax + 2);
    int k = m;
    if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1) ||
	(k == 0) || (k == kmax + 1)) {
      return 0.0;
    }
    return 1.0;
  }
} gen_f;

void make_vector(double *ap, double *ae, double *aw, double *an, double *as, double *ab, double *at,
		 sketo::dist_list<double> *b, sketo::dist_list<double> *f)
{
  *b = sl::generate(nmax_b, gen_b);
  *f = sl::generate(nmax_b, gen_f);

  *ap = 2.0;
  *ae = *aw = *an = *as = *at = *ab = 1.0;
}

void probmv(double ap, double ae, double aw, double an, double as, double at, double ab,
	    const sketo::dist_list<double> &b, sketo::dist_list<double> *c,
	    const sketo::dist_list<double> &f)
{
  //  const int dx = 1;
  const int dy = imax + 2;
  const int dz = (imax + 2) * (jmax + 2);

  sketo::dist_list<double> bs = gshift(dy, 0.0, b);
  sketo::dist_list<double> bn = gshift(-dy, 0.0, b);
  sketo::dist_list<double> bb = gshift(dz, 0.0, b);
  sketo::dist_list<double> bt = gshift(-dz, 0.0, b);
  
  *c = sl::zipwith(std::multiplies<double>(), f,
           sl::zipwith(std::plus<double>(),
           sl::zipwith(std::plus<double>(),
           sl::zipwith(std::plus<double>(),
           sl::zipwith(std::plus<double>(),
           sl::zipwith(std::plus<double>(),
           sl::zipwith(std::plus<double>(),
               sl::map(std::bind1st(std::multiplies<double>(), ap), b),
               sl::map(std::bind1st(std::multiplies<double>(), aw), sl::shiftr(0.0, b))),
               sl::map(std::bind1st(std::multiplies<double>(), ae), sl::shiftl(0.0, b))),
               sl::map(std::bind1st(std::multiplies<double>(), as), bs)),
               sl::map(std::bind1st(std::multiplies<double>(), an), bn)),
               sl::map(std::bind1st(std::multiplies<double>(), ab), bb)),
               sl::map(std::bind1st(std::multiplies<double>(), at), bt)));

}

std::pair<int, double>
bicgstab_band(double ap, double ae, double aw ,double an, double as, double at, double ab,
	      sketo::dist_list<double> *x,
	      sketo::dist_list<double> *b,
	      const sketo::dist_list<double> &f)
{
//   double er0 = EPS;

  sketo::dist_list<double> c, r, p, r0, e, y, v;

  probmv(ap, ae, aw, an, as, at, ab, *x, &c, f);

  r = sl::zipwith(std::minus<double>(), *b, c);
  double c1 = sl::reduce(std::plus<double>(),
			 sl::map(sketo::functions::square<double>(), r));

//   if (c1 < er0) {
//     return std::pair<int, double>(0, c1);
//   }

  p = r.clone();
  r0 = r.clone();

  std::pair<int, double> ret;
  for (int it = 0; it < ITRMAX; it++) {

    probmv(ap, ae, aw, an, as, at, ab, p, &y, f);

    double c2 = sl::reduce(std::plus<double>(),
			   sl::zipwith(std::multiplies<double>(), r0, y));
    double alp = c1 / c2;

    e = sl::zipwith(std::minus<double>(),
		    r,
		    sl::map(std::bind1st(std::multiplies<double>(), alp), y));

    probmv(ap, ae, aw, an, as, at, ab, e, &v, f);

    double ev = sl::reduce(std::plus<double>(),
			   sl::zipwith(std::multiplies<double>(), e, v));
    double vv = sl::reduce(std::plus<double>(),
			   sl::zipwith(std::multiplies<double>(), v, v));
    
    double c3 = ev / vv;
    *x = sl::zipwith(std::plus<double>(),
		     *x, sl::zipwith(std::plus<double>(),
				     sl::map(std::bind1st(std::multiplies<double>(), alp), p),
				     sl::map(std::bind1st(std::multiplies<double>(), c3), e)));
    r = sl::zipwith(std::minus<double>(), e, sl::map(std::bind1st(std::multiplies<double>(), c3), v));
    
    double rr = sl::reduce(std::plus<double>(),
			   sl::zipwith(std::multiplies<double>(), r, r));
//     if (std::fabs(rr) < er0) { ret = std::pair<int, double>(it + 1, rr); break; }

    c1 = sl::reduce(std::plus<double>(),
		    sl::zipwith(std::multiplies<double>(), r0, r));
    double bet = c1/(c2 * c3);
    p = sl::zipwith(std::plus<double>(),
		    r, sl::zipwith(std::plus<double>(),
				   sl::map(std::bind1st(std::multiplies<double>(), bet), p),
				   sl::map(std::bind1st(std::multiplies<double>(), -bet * c3), y)));
    
    ret = std::pair<int, double>(it + 1, rr);
  }
  return ret;
}
  
struct gen_x_t : public sketo::functions::base<double (int)> {
  double operator()(int m) const {
    int i = m % (imax + 2); m /= (imax + 2);
    int j = m % (jmax + 2); m /= (jmax + 2);
    int k = m;
    if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1) ||
	(k == 0) || (k == kmax + 1)) {
      return 0.0;
    }
    return 1.0 + 0.01;
  }
} gen_x;

int sketo::main(int, char**)
{
  double ap, ae, aw, an, as, at, ab;
  sketo::dist_list<double> b, f;
  make_vector(&ap, &ae, &aw, &an, &as, &at, &ab, &b, &f);

  sketo::dist_list<double> x = sl::generate(nmax_b, gen_x);
  
  double st, et;
  MPI_Barrier(MPI_COMM_WORLD);
  st = MPI_Wtime();
#ifndef PROBMV_ONLY
  std::pair<int, double> result = bicgstab_band(ap, ae, aw, an, as, at, ab,
						&x, &b, f);
  std::cout << "error = " << result.second << std::endl;
#else
  ap = -4.5;
  sketo::dist_list<double> y(nmax_b);
  // avoid too-much optimization
  double tmpval = 0.0;
  for (int i = 0; i < ITRMAX * 2; i++) {
    probmv(ap, ae, aw, an, as, at, ab, x, &y, f);
    tmpval += sketo::_impl::dist_list_buffer<double>::get_buffer(y)->buf[0];
  }
  sketo::cout << tmpval << std::endl;
#endif
  MPI_Barrier(MPI_COMM_WORLD);
  et = MPI_Wtime();
  sketo::cout << "time = " << (et - st) << std::endl;

  return 0;
}



