template <typename A>
sketo::dist_list<A>
gshift(int diff, const A& default_value, const sketo::dist_list<A> &src)
{
  MPI_Barrier(MPI_COMM_WORLD);
  //   if (rank == 0) cout << "shifting by " << diff << endl;
  sketo::dist_list<A> dst(src.get_global_size());
  
  int local_send_begin = -1;
  int local_send_end = -1;
  
  MPI_Request** send_req = new MPI_Request*[sketo::procs];
  MPI_Request** recv_req = new MPI_Request*[sketo::procs];
  for (int i = 0; i < sketo::procs; i++) {
    send_req[i] = NULL; recv_req[i] = NULL;
  }
  
  // computing send
  int start_index = 0;
  int pd = 0;
  
  int local_begin, local_size;
  sketo::_impl::dist_list_buffer<void>::compute_local_array(src.get_global_size(), sketo::procs, sketo::rank, &local_begin, &local_size);
  if (local_begin + diff < 0) start_index = - diff - local_begin;
  while(start_index < local_size && pd < sketo::procs) {
    int pd_begin, pd_size;
    sketo::_impl::dist_list_buffer<void>::compute_local_array(src.get_global_size(), sketo::procs, pd, &pd_begin, &pd_size);
    int target_begin = local_begin + start_index + diff;
    if (pd_begin <= target_begin && target_begin < pd_begin + pd_size) {
      int src_end = std::min(pd_begin + pd_size - diff, local_begin + local_size) - local_begin;
      //       cout << "Proc[" << rank << "]->Proc[" << pd << "] data[" << local_begin + start_index << "," << src_end << "]" << endl;
      // send
      if (pd == sketo::rank) {
	local_send_begin = start_index;
	local_send_end = src_end;
      } else {
	send_req[pd] = new MPI_Request;
	MPI_Isend(sketo::_impl::dist_list_buffer<A>::get_buffer(src)->buf + start_index, (src_end - start_index) * sizeof(A), MPI_BYTE,
		  pd, 0, MPI_COMM_WORLD, send_req[pd]);
      }
      start_index = src_end;
    }
    pd++;
  }
  
  // computing recv
  start_index = 0;
  int ps = 0;
  if (local_begin - diff < 0) {
    start_index = diff - local_begin;
    start_index = std::min<int>(start_index, local_size);
    for (int i = 0; i < start_index; i++) {
      sketo::_impl::dist_list_buffer<A>::get_buffer(dst)->buf[i] = default_value;
    }
  }
  while(start_index < local_size && ps < sketo::procs) {
    int ps_begin, ps_size;
    sketo::_impl::dist_list_buffer<void>::compute_local_array(src.get_global_size(), sketo::procs, ps, &ps_begin, &ps_size);
    int target_begin = local_begin + start_index - diff;
    if (ps_begin <= target_begin && target_begin < ps_begin + ps_size) {
      int dest_end = std::min(ps_begin + ps_size + diff, local_begin + local_size) - local_begin;
      //       cout << "Proc[" << rank << "]->Proc[" << ps << "] data[" << local_begin + start_index << "," << dest_end << "]" << endl;
      // recv
      if (ps == sketo::rank) {
      } else {
	recv_req[ps] = new MPI_Request;
	MPI_Irecv(sketo::_impl::dist_list_buffer<A>::get_buffer(dst)->buf + start_index, (dest_end - start_index) * sizeof(A), MPI_BYTE,
		  ps, 0, MPI_COMM_WORLD, recv_req[ps]);
      }
      start_index = dest_end;
    }
    ps++;
  }
  if (src.get_global_size() + diff < local_begin + local_size) {
    for (int i = src.get_global_size() + diff - local_begin; i < local_size; i++) {
      sketo::_impl::dist_list_buffer<A>::get_buffer(dst)->buf[i] = default_value;
    }
  }
  
  for (int i = local_send_begin; i < local_send_end; i++) {
    sketo::_impl::dist_list_buffer<A>::get_buffer(dst)->buf[i + diff] = sketo::_impl::dist_list_buffer<A>::get_buffer(src)->buf[i];
  }
  
  // wait
  for (int p = 0; p < sketo::procs; ++p) {
    if (send_req[p]) MPI_Wait(send_req[p], MPI_STATUS_IGNORE);
    if (recv_req[p]) MPI_Wait(recv_req[p], MPI_STATUS_IGNORE);
  }
  
  delete[] send_req;
  delete[] recv_req;
  
  return dst;
}
